package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.ast.Invocation;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.ImportBinding;
import org.eclipse.jdt.internal.compiler.lookup.ImportConflictBinding;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.LocalTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.MissingTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemFieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeCollisionException;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.CompoundNameVector;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.util.HashtableOfType;
import org.eclipse.jdt.internal.compiler.util.ObjectVector;
import org.eclipse.jdt.internal.compiler.util.SimpleNameVector;

public class CompilationUnitScope extends Scope {
  public LookupEnvironment environment;
  
  public CompilationUnitDeclaration referenceContext;
  
  public char[][] currentPackageName;
  
  public PackageBinding fPackage;
  
  public ImportBinding[] imports;
  
  public int importPtr;
  
  public HashtableOfObject typeOrPackageCache;
  
  public SourceTypeBinding[] topLevelTypes;
  
  private CompoundNameVector qualifiedReferences;
  
  private SimpleNameVector simpleNameReferences;
  
  private SimpleNameVector rootReferences;
  
  private ObjectVector referencedTypes;
  
  private ObjectVector referencedSuperTypes;
  
  HashtableOfType constantPoolNameUsage;
  
  private int captureID = 1;
  
  private ImportBinding[] tempImports;
  
  public boolean suppressImportErrors;
  
  private boolean skipCachingImports;
  
  boolean connectingHierarchy;
  
  private ArrayList<Invocation> inferredInvocations;
  
  Map<InferenceVariable.InferenceVarKey, InferenceVariable> uniqueInferenceVariables = new HashMap<>();
  
  public CompilationUnitScope(CompilationUnitDeclaration unit, LookupEnvironment environment) {
    super(4, null);
    this.environment = environment;
    this.referenceContext = unit;
    unit.scope = this;
    this.currentPackageName = (unit.currentPackage == null) ? CharOperation.NO_CHAR_CHAR : unit.currentPackage.tokens;
    if ((compilerOptions()).produceReferenceInfo) {
      this.qualifiedReferences = new CompoundNameVector();
      this.simpleNameReferences = new SimpleNameVector();
      this.rootReferences = new SimpleNameVector();
      this.referencedTypes = new ObjectVector();
      this.referencedSuperTypes = new ObjectVector();
    } else {
      this.qualifiedReferences = null;
      this.simpleNameReferences = null;
      this.rootReferences = null;
      this.referencedTypes = null;
      this.referencedSuperTypes = null;
    } 
  }
  
  void buildFieldsAndMethods() {
    for (int i = 0, length = this.topLevelTypes.length; i < length; i++)
      (this.topLevelTypes[i]).scope.buildFieldsAndMethods(); 
  }
  
  void buildTypeBindings(AccessRestriction accessRestriction) {
    this.topLevelTypes = new SourceTypeBinding[0];
    boolean firstIsSynthetic = false;
    if (this.referenceContext.compilationResult.compilationUnit != null) {
      char[][] expectedPackageName = this.referenceContext.compilationResult.compilationUnit.getPackageName();
      if (expectedPackageName != null && 
        !CharOperation.equals(this.currentPackageName, expectedPackageName)) {
        if (this.referenceContext.currentPackage != null || 
          this.referenceContext.types != null || 
          this.referenceContext.imports != null)
          problemReporter().packageIsNotExpectedPackage(this.referenceContext); 
        this.currentPackageName = (expectedPackageName.length == 0) ? CharOperation.NO_CHAR_CHAR : expectedPackageName;
      } 
    } 
    if (this.currentPackageName == CharOperation.NO_CHAR_CHAR) {
      this.fPackage = this.environment.defaultPackage;
    } else {
      if ((this.fPackage = this.environment.createPackage(this.currentPackageName)) == null) {
        if (this.referenceContext.currentPackage != null)
          problemReporter().packageCollidesWithType(this.referenceContext); 
        this.fPackage = this.environment.defaultPackage;
        return;
      } 
      if (this.referenceContext.isPackageInfo()) {
        if (this.referenceContext.types == null || this.referenceContext.types.length == 0) {
          this.referenceContext.types = new TypeDeclaration[1];
          this.referenceContext.createPackageInfoType();
          firstIsSynthetic = true;
        } 
        if (this.referenceContext.currentPackage != null && this.referenceContext.currentPackage.annotations != null)
          (this.referenceContext.types[0]).annotations = this.referenceContext.currentPackage.annotations; 
      } 
      recordQualifiedReference(this.currentPackageName);
    } 
    TypeDeclaration[] types = this.referenceContext.types;
    int typeLength = (types == null) ? 0 : types.length;
    this.topLevelTypes = new SourceTypeBinding[typeLength];
    int count = 0;
    for (int i = 0; i < typeLength; i++) {
      TypeDeclaration typeDecl = types[i];
      if (this.environment.isProcessingAnnotations && this.environment.isMissingType(typeDecl.name))
        throw new SourceTypeCollisionException(); 
      ReferenceBinding typeBinding = this.fPackage.getType0(typeDecl.name);
      recordSimpleReference(typeDecl.name);
      if (typeBinding != null && typeBinding.isValidBinding() && !(typeBinding instanceof org.eclipse.jdt.internal.compiler.lookup.UnresolvedReferenceBinding)) {
        if (this.environment.isProcessingAnnotations)
          throw new SourceTypeCollisionException(); 
        problemReporter().duplicateTypes(this.referenceContext, typeDecl);
      } else {
        if (this.fPackage != this.environment.defaultPackage && this.fPackage.getPackage(typeDecl.name) != null)
          problemReporter().typeCollidesWithPackage(this.referenceContext, typeDecl); 
        if ((typeDecl.modifiers & 0x1) != 0) {
          char[] mainTypeName;
          if ((mainTypeName = this.referenceContext.getMainTypeName()) != null && 
            !CharOperation.equals(mainTypeName, typeDecl.name))
            problemReporter().publicClassMustMatchFileName(this.referenceContext, typeDecl); 
        } 
        ClassScope child = new ClassScope(this, typeDecl);
        SourceTypeBinding type = child.buildType(null, this.fPackage, accessRestriction);
        if (firstIsSynthetic && i == 0)
          type.modifiers |= 0x1000; 
        if (type != null)
          this.topLevelTypes[count++] = type; 
      } 
    } 
    if (count != this.topLevelTypes.length)
      System.arraycopy(this.topLevelTypes, 0, this.topLevelTypes = new SourceTypeBinding[count], 0, count); 
  }
  
  void checkAndSetImports() {
    if (this.referenceContext.imports == null) {
      this.imports = getDefaultImports();
      return;
    } 
    int numberOfStatements = this.referenceContext.imports.length;
    int numberOfImports = numberOfStatements + 1;
    for (int i = 0; i < numberOfStatements; i++) {
      ImportReference importReference = this.referenceContext.imports[i];
      if ((importReference.bits & 0x20000) != 0 && CharOperation.equals(TypeConstants.JAVA_LANG, importReference.tokens) && !importReference.isStatic()) {
        numberOfImports--;
        break;
      } 
    } 
    ImportBinding[] resolvedImports = new ImportBinding[numberOfImports];
    resolvedImports[0] = getDefaultImports()[0];
    int index = 1;
    for (int j = 0; j < numberOfStatements; j++) {
      ImportReference importReference = this.referenceContext.imports[j];
      char[][] compoundName = importReference.tokens;
      for (int k = 0; k < index; k++) {
        ImportBinding resolved = resolvedImports[k];
        if (resolved.onDemand == (((importReference.bits & 0x20000) != 0)) && resolved.isStatic() == importReference.isStatic() && 
          CharOperation.equals(compoundName, (resolvedImports[k]).compoundName))
          // Byte code: goto -> 325 
      } 
      if ((importReference.bits & 0x20000) != 0) {
        if (!CharOperation.equals(compoundName, this.currentPackageName)) {
          Binding importBinding = findImport(compoundName, compoundName.length);
          if (importBinding.isValidBinding() && (!importReference.isStatic() || !(importBinding instanceof PackageBinding)))
            resolvedImports[index++] = new ImportBinding(compoundName, true, importBinding, importReference); 
        } 
      } else {
        resolvedImports[index++] = new ImportBinding(compoundName, false, null, importReference);
      } 
    } 
    if (resolvedImports.length > index)
      System.arraycopy(resolvedImports, 0, resolvedImports = new ImportBinding[index], 0, index); 
    this.imports = resolvedImports;
  }
  
  void checkParameterizedTypes() {
    if ((compilerOptions()).sourceLevel < 3211264L)
      return; 
    for (int i = 0, length = this.topLevelTypes.length; i < length; i++) {
      ClassScope scope = (this.topLevelTypes[i]).scope;
      scope.checkParameterizedTypeBounds();
      scope.checkParameterizedSuperTypeCollisions();
    } 
  }
  
  public char[] computeConstantPoolName(LocalTypeBinding localType) {
    char[] candidateName;
    if (localType.constantPoolName != null)
      return localType.constantPoolName; 
    if (this.constantPoolNameUsage == null)
      this.constantPoolNameUsage = new HashtableOfType(); 
    SourceTypeBinding sourceTypeBinding = localType.scope.outerMostClassScope().enclosingSourceType();
    int index = 0;
    boolean isCompliant15 = ((compilerOptions()).complianceLevel >= 3211264L);
    while (true) {
      if (localType.isMemberType()) {
        if (index == 0) {
          candidateName = CharOperation.concat(
              localType.enclosingType().constantPoolName(), 
              localType.sourceName, 
              '$');
        } else {
          candidateName = CharOperation.concat(
              localType.enclosingType().constantPoolName(), 
              '$', 
              String.valueOf(index).toCharArray(), 
              '$', 
              localType.sourceName);
        } 
      } else if (localType.isAnonymousType()) {
        if (isCompliant15) {
          candidateName = CharOperation.concat(
              localType.enclosingType.constantPoolName(), 
              String.valueOf(index + 1).toCharArray(), 
              '$');
        } else {
          candidateName = CharOperation.concat(
              sourceTypeBinding.constantPoolName(), 
              String.valueOf(index + 1).toCharArray(), 
              '$');
        } 
      } else if (isCompliant15) {
        candidateName = CharOperation.concat(
            CharOperation.concat(
              localType.enclosingType().constantPoolName(), 
              String.valueOf(index + 1).toCharArray(), 
              '$'), 
            localType.sourceName);
      } else {
        candidateName = CharOperation.concat(
            sourceTypeBinding.constantPoolName(), 
            '$', 
            String.valueOf(index + 1).toCharArray(), 
            '$', 
            localType.sourceName);
      } 
      if (this.constantPoolNameUsage.get(candidateName) != null) {
        index++;
        continue;
      } 
      break;
    } 
    this.constantPoolNameUsage.put(candidateName, (ReferenceBinding)localType);
    return candidateName;
  }
  
  void connectTypeHierarchy() {
    for (int i = 0, length = this.topLevelTypes.length; i < length; i++)
      (this.topLevelTypes[i]).scope.connectTypeHierarchy(); 
  }
  
  void faultInImports() {
    boolean unresolvedFound = false;
    boolean reportUnresolved = !this.suppressImportErrors;
    if (this.typeOrPackageCache != null && !this.skipCachingImports)
      return; 
    if (this.referenceContext.imports == null) {
      this.typeOrPackageCache = new HashtableOfObject(1);
      return;
    } 
    int numberOfStatements = this.referenceContext.imports.length;
    HashtableOfType typesBySimpleNames = null;
    for (int i = 0; i < numberOfStatements; i++) {
      if (((this.referenceContext.imports[i]).bits & 0x20000) == 0) {
        typesBySimpleNames = new HashtableOfType(this.topLevelTypes.length + numberOfStatements);
        for (int m = 0, n = this.topLevelTypes.length; m < n; m++)
          typesBySimpleNames.put((this.topLevelTypes[m]).sourceName, (ReferenceBinding)this.topLevelTypes[m]); 
        break;
      } 
    } 
    int numberOfImports = numberOfStatements + 1;
    int j;
    for (j = 0; j < numberOfStatements; j++) {
      ImportReference importReference = this.referenceContext.imports[j];
      if ((importReference.bits & 0x20000) != 0 && CharOperation.equals(TypeConstants.JAVA_LANG, importReference.tokens) && !importReference.isStatic()) {
        numberOfImports--;
        break;
      } 
    } 
    this.tempImports = new ImportBinding[numberOfImports];
    this.tempImports[0] = getDefaultImports()[0];
    this.importPtr = 1;
    for (j = 0; j < numberOfStatements; j++) {
      ImportReference importReference = this.referenceContext.imports[j];
      char[][] compoundName = importReference.tokens;
      for (int m = 0; m < this.importPtr; m++) {
        ImportBinding resolved = this.tempImports[m];
        if (resolved.onDemand == (((importReference.bits & 0x20000) != 0)) && resolved.isStatic() == importReference.isStatic() && 
          CharOperation.equals(compoundName, resolved.compoundName)) {
          problemReporter().unusedImport(importReference);
          // Byte code: goto -> 640
        } 
      } 
      if ((importReference.bits & 0x20000) != 0) {
        if (CharOperation.equals(compoundName, this.currentPackageName)) {
          problemReporter().unusedImport(importReference);
        } else {
          Binding importBinding = findImport(compoundName, compoundName.length);
          if (!importBinding.isValidBinding()) {
            problemReporter().importProblem(importReference, importBinding);
          } else if (importReference.isStatic() && importBinding instanceof PackageBinding) {
            problemReporter().cannotImportPackage(importReference);
          } else {
            recordImportBinding(new ImportBinding(compoundName, true, importBinding, importReference));
          } 
        } 
      } else {
        Binding importBinding = findSingleImport(compoundName, 13, importReference.isStatic());
        if (!importBinding.isValidBinding() && 
          importBinding.problemId() != 3) {
          unresolvedFound = true;
          if (reportUnresolved)
            problemReporter().importProblem(importReference, importBinding); 
        } else if (importBinding instanceof PackageBinding) {
          problemReporter().cannotImportPackage(importReference);
        } else if (checkAndRecordImportBinding(importBinding, typesBySimpleNames, importReference, compoundName) != -1) {
          if (importReference.isStatic())
            if (importBinding.kind() == 1) {
              checkMoreStaticBindings(compoundName, typesBySimpleNames, 12, importReference);
            } else if (importBinding.kind() == 8) {
              checkMoreStaticBindings(compoundName, typesBySimpleNames, 4, importReference);
            }  
        } 
      } 
    } 
    if (this.tempImports.length > this.importPtr)
      System.arraycopy(this.tempImports, 0, this.tempImports = new ImportBinding[this.importPtr], 0, this.importPtr); 
    this.imports = this.tempImports;
    int length = this.imports.length;
    this.typeOrPackageCache = new HashtableOfObject(length);
    for (int k = 0; k < length; k++) {
      ImportBinding binding = this.imports[k];
      if ((!binding.onDemand && binding.resolvedImport instanceof ReferenceBinding) || binding instanceof ImportConflictBinding)
        this.typeOrPackageCache.put(binding.compoundName[binding.compoundName.length - 1], binding); 
    } 
    this.skipCachingImports = (this.suppressImportErrors && unresolvedFound);
  }
  
  public void faultInTypes() {
    faultInImports();
    for (int i = 0, length = this.topLevelTypes.length; i < length; i++)
      this.topLevelTypes[i].faultInTypesForFieldsAndMethods(); 
  }
  
  public Binding findImport(char[][] compoundName, boolean findStaticImports, boolean onDemand) {
    if (onDemand)
      return findImport(compoundName, compoundName.length); 
    return findSingleImport(compoundName, 13, findStaticImports);
  }
  
  private Binding findImport(char[][] compoundName, int length) {
    Binding binding;
    ReferenceBinding type;
    recordQualifiedReference(compoundName);
    PackageBinding packageBinding = this.environment.getTopLevelPackage(compoundName[0]);
    int i = 1;
    if (packageBinding != null) {
      PackageBinding packageBinding1 = packageBinding;
      while (true) {
        if (i >= length)
          return (Binding)packageBinding1; 
        binding = packageBinding1.getTypeOrPackage(compoundName[i++]);
        if (binding == null || !binding.isValidBinding()) {
          binding = null;
          break;
        } 
        if (!(binding instanceof PackageBinding))
          break; 
        packageBinding1 = (PackageBinding)binding;
      } 
    } 
    if (binding == null) {
      if ((compilerOptions()).complianceLevel >= 3145728L)
        return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), null, 1); 
      type = findType(compoundName[0], this.environment.defaultPackage, this.environment.defaultPackage);
      if (type == null || !type.isValidBinding())
        return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), null, 1); 
      i = 1;
    } else {
      type = (ReferenceBinding)binding;
    } 
    while (i < length) {
      type = (ReferenceBinding)this.environment.convertToRawType((TypeBinding)type, false);
      if (!type.canBeSeenBy(this.fPackage))
        return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), type, 2); 
      char[] name = compoundName[i++];
      type = type.getMemberType(name);
      if (type == null)
        return (Binding)new ProblemReferenceBinding(CharOperation.subarray(compoundName, 0, i), null, 1); 
    } 
    if (!type.canBeSeenBy(this.fPackage))
      return (Binding)new ProblemReferenceBinding(compoundName, type, 2); 
    return (Binding)type;
  }
  
  private Binding findSingleImport(char[][] compoundName, int mask, boolean findStaticImports) {
    if (compoundName.length == 1) {
      if ((compilerOptions()).complianceLevel >= 3145728L)
        return (Binding)new ProblemReferenceBinding(compoundName, null, 1); 
      ReferenceBinding typeBinding = findType(compoundName[0], this.environment.defaultPackage, this.fPackage);
      if (typeBinding == null)
        return (Binding)new ProblemReferenceBinding(compoundName, null, 1); 
      return (Binding)typeBinding;
    } 
    if (findStaticImports)
      return findSingleStaticImport(compoundName, mask); 
    return findImport(compoundName, compoundName.length);
  }
  
  private Binding findSingleStaticImport(char[][] compoundName, int mask) {
    Binding binding = findImport(compoundName, compoundName.length - 1);
    if (!binding.isValidBinding())
      return binding; 
    char[] name = compoundName[compoundName.length - 1];
    if (binding instanceof PackageBinding) {
      Binding temp = ((PackageBinding)binding).getTypeOrPackage(name);
      if (temp != null && temp instanceof ReferenceBinding)
        return (Binding)new ProblemReferenceBinding(compoundName, (ReferenceBinding)temp, 14); 
      return binding;
    } 
    ReferenceBinding type = (ReferenceBinding)binding;
    FieldBinding field = ((mask & 0x1) != 0) ? findField((TypeBinding)type, name, null, true) : null;
    if (field != null) {
      if (field.problemId() == 3 && ((ProblemFieldBinding)field).closestMatch.isStatic())
        return (Binding)field; 
      if (field.isValidBinding() && field.isStatic() && field.canBeSeenBy((TypeBinding)type, null, this))
        return (Binding)field; 
    } 
    MethodBinding method = ((mask & 0x8) != 0) ? findStaticMethod(type, name) : null;
    if (method != null)
      return (Binding)method; 
    type = findMemberType(name, type);
    if (type == null || !type.isStatic()) {
      if (field != null && !field.isValidBinding() && field.problemId() != 1)
        return (Binding)field; 
      return (Binding)new ProblemReferenceBinding(compoundName, type, 1);
    } 
    if (type.isValidBinding() && !type.canBeSeenBy(this.fPackage))
      return (Binding)new ProblemReferenceBinding(compoundName, type, 2); 
    if (type.problemId() == 2)
      return (Binding)new ProblemReferenceBinding(compoundName, ((ProblemReferenceBinding)type).closestMatch, 2); 
    return (Binding)type;
  }
  
  private MethodBinding findStaticMethod(ReferenceBinding currentType, char[] selector) {
    if (!currentType.canBeSeenBy(this))
      return null; 
    while (true) {
      currentType.initializeForStaticImports();
      MethodBinding[] methods = currentType.getMethods(selector);
      if (methods != Binding.NO_METHODS)
        for (int i = methods.length; --i >= 0; ) {
          MethodBinding method = methods[i];
          if (method.isStatic() && method.canBeSeenBy(this.fPackage))
            return method; 
        }  
      if ((currentType = currentType.superclass()) == null)
        return null; 
    } 
  }
  
  ImportBinding[] getDefaultImports() {
    Binding binding;
    PackageBinding packageBinding1;
    if (this.environment.defaultImports != null)
      return this.environment.defaultImports; 
    PackageBinding packageBinding2 = this.environment.getTopLevelPackage(TypeConstants.JAVA);
    if (packageBinding2 != null)
      binding = packageBinding2.getTypeOrPackage(TypeConstants.JAVA_LANG[1]); 
    if (binding == null || !binding.isValidBinding()) {
      problemReporter().isClassPathCorrect(
          TypeConstants.JAVA_LANG_OBJECT, 
          this.referenceContext, 
          this.environment.missingClassFileLocation);
      MissingTypeBinding missingTypeBinding = this.environment.createMissingType(null, TypeConstants.JAVA_LANG_OBJECT);
      packageBinding1 = ((BinaryTypeBinding)missingTypeBinding).fPackage;
    } 
    this.environment.defaultImports = new ImportBinding[] { new ImportBinding(TypeConstants.JAVA_LANG, true, (Binding)packageBinding1, null) };
    return new ImportBinding[1];
  }
  
  public final Binding getImport(char[][] compoundName, boolean onDemand, boolean isStaticImport) {
    if (onDemand)
      return findImport(compoundName, compoundName.length); 
    return findSingleImport(compoundName, 13, isStaticImport);
  }
  
  public int nextCaptureID() {
    return this.captureID++;
  }
  
  public ProblemReporter problemReporter() {
    ProblemReporter problemReporter = this.referenceContext.problemReporter;
    problemReporter.referenceContext = (ReferenceContext)this.referenceContext;
    return problemReporter;
  }
  
  void recordQualifiedReference(char[][] qualifiedName) {
    if (this.qualifiedReferences == null)
      return; 
    int length = qualifiedName.length;
    if (length > 1) {
      recordRootReference(qualifiedName[0]);
      while (!this.qualifiedReferences.contains(qualifiedName)) {
        this.qualifiedReferences.add(qualifiedName);
        if (length == 2) {
          recordSimpleReference(qualifiedName[0]);
          recordSimpleReference(qualifiedName[1]);
          return;
        } 
        length--;
        recordSimpleReference(qualifiedName[length]);
        System.arraycopy(qualifiedName, 0, qualifiedName = new char[length][], 0, length);
      } 
    } else if (length == 1) {
      recordRootReference(qualifiedName[0]);
      recordSimpleReference(qualifiedName[0]);
    } 
  }
  
  void recordReference(char[][] qualifiedEnclosingName, char[] simpleName) {
    recordQualifiedReference(qualifiedEnclosingName);
    if (qualifiedEnclosingName.length == 0)
      recordRootReference(simpleName); 
    recordSimpleReference(simpleName);
  }
  
  void recordReference(ReferenceBinding type, char[] simpleName) {
    ReferenceBinding actualType = typeToRecord((TypeBinding)type);
    if (actualType != null)
      recordReference(actualType.compoundName, simpleName); 
  }
  
  void recordRootReference(char[] simpleName) {
    if (this.rootReferences == null)
      return; 
    if (!this.rootReferences.contains(simpleName))
      this.rootReferences.add(simpleName); 
  }
  
  void recordSimpleReference(char[] simpleName) {
    if (this.simpleNameReferences == null)
      return; 
    if (!this.simpleNameReferences.contains(simpleName))
      this.simpleNameReferences.add(simpleName); 
  }
  
  void recordSuperTypeReference(TypeBinding type) {
    if (this.referencedSuperTypes == null)
      return; 
    ReferenceBinding actualType = typeToRecord(type);
    if (actualType != null && !this.referencedSuperTypes.containsIdentical(actualType))
      this.referencedSuperTypes.add(actualType); 
  }
  
  public void recordTypeConversion(TypeBinding superType, TypeBinding subType) {
    recordSuperTypeReference(subType);
  }
  
  void recordTypeReference(TypeBinding type) {
    if (this.referencedTypes == null)
      return; 
    ReferenceBinding actualType = typeToRecord(type);
    if (actualType != null && !this.referencedTypes.containsIdentical(actualType))
      this.referencedTypes.add(actualType); 
  }
  
  void recordTypeReferences(TypeBinding[] types) {
    if (this.referencedTypes == null)
      return; 
    if (types == null || types.length == 0)
      return; 
    for (int i = 0, max = types.length; i < max; i++) {
      ReferenceBinding actualType = typeToRecord(types[i]);
      if (actualType != null && !this.referencedTypes.containsIdentical(actualType))
        this.referencedTypes.add(actualType); 
    } 
  }
  
  Binding resolveSingleImport(ImportBinding importBinding, int mask) {
    if (importBinding.resolvedImport == null) {
      importBinding.resolvedImport = findSingleImport(importBinding.compoundName, mask, importBinding.isStatic());
      if (!importBinding.resolvedImport.isValidBinding() || importBinding.resolvedImport instanceof PackageBinding) {
        if (importBinding.resolvedImport.problemId() == 3)
          return importBinding.resolvedImport; 
        if (this.imports != null) {
          ImportBinding[] newImports = new ImportBinding[this.imports.length - 1];
          for (int i = 0, n = 0, max = this.imports.length; i < max; i++) {
            if (this.imports[i] != importBinding)
              newImports[n++] = this.imports[i]; 
          } 
          this.imports = newImports;
        } 
        return null;
      } 
    } 
    return importBinding.resolvedImport;
  }
  
  public void storeDependencyInfo() {
    int i;
    for (i = 0; i < this.referencedSuperTypes.size; i++) {
      ReferenceBinding type = (ReferenceBinding)this.referencedSuperTypes.elementAt(i);
      if (!this.referencedTypes.containsIdentical(type))
        this.referencedTypes.add(type); 
      if (!type.isLocalType()) {
        ReferenceBinding enclosing = type.enclosingType();
        if (enclosing != null)
          recordSuperTypeReference((TypeBinding)enclosing); 
      } 
      ReferenceBinding superclass = type.superclass();
      if (superclass != null)
        recordSuperTypeReference((TypeBinding)superclass); 
      ReferenceBinding[] interfaces = type.superInterfaces();
      if (interfaces != null)
        for (int n = 0, length = interfaces.length; n < length; n++)
          recordSuperTypeReference((TypeBinding)interfaces[n]);  
    } 
    int l;
    for (i = 0, l = this.referencedTypes.size; i < l; i++) {
      ReferenceBinding type = (ReferenceBinding)this.referencedTypes.elementAt(i);
      if (!type.isLocalType())
        recordQualifiedReference(type.isMemberType() ? 
            CharOperation.splitOn('.', type.readableName()) : 
            type.compoundName); 
    } 
    int size = this.qualifiedReferences.size;
    char[][][] qualifiedRefs = new char[size][][];
    for (int j = 0; j < size; j++)
      qualifiedRefs[j] = this.qualifiedReferences.elementAt(j); 
    this.referenceContext.compilationResult.qualifiedReferences = qualifiedRefs;
    size = this.simpleNameReferences.size;
    char[][] simpleRefs = new char[size][];
    for (int k = 0; k < size; k++)
      simpleRefs[k] = this.simpleNameReferences.elementAt(k); 
    this.referenceContext.compilationResult.simpleNameReferences = simpleRefs;
    size = this.rootReferences.size;
    char[][] rootRefs = new char[size][];
    for (int m = 0; m < size; m++)
      rootRefs[m] = this.rootReferences.elementAt(m); 
    this.referenceContext.compilationResult.rootReferences = rootRefs;
  }
  
  public String toString() {
    return "--- CompilationUnit Scope : " + new String(this.referenceContext.getFileName());
  }
  
  private ReferenceBinding typeToRecord(TypeBinding type) {
    if (type == null)
      return null; 
    while (type.isArrayType())
      type = ((ArrayBinding)type).leafComponentType(); 
    switch (type.kind()) {
      case 132:
      case 516:
      case 4100:
      case 8196:
      case 32772:
      case 65540:
        return null;
      case 260:
      case 1028:
        type = type.erasure();
        break;
    } 
    ReferenceBinding refType = (ReferenceBinding)type;
    if (refType.isLocalType())
      return null; 
    return refType;
  }
  
  public void verifyMethods(MethodVerifier verifier) {
    for (int i = 0, length = this.topLevelTypes.length; i < length; i++)
      this.topLevelTypes[i].verifyMethods(verifier); 
  }
  
  private void recordImportBinding(ImportBinding bindingToAdd) {
    if (this.tempImports.length == this.importPtr)
      System.arraycopy(this.tempImports, 0, this.tempImports = new ImportBinding[this.importPtr + 1], 0, this.importPtr); 
    this.tempImports[this.importPtr++] = bindingToAdd;
  }
  
  private void checkMoreStaticBindings(char[][] compoundName, HashtableOfType typesBySimpleNames, int mask, ImportReference importReference) {
    Binding importBinding = findSingleStaticImport(compoundName, mask);
    if (!importBinding.isValidBinding()) {
      if (importBinding.problemId() == 3)
        checkAndRecordImportBinding(importBinding, typesBySimpleNames, importReference, compoundName); 
    } else {
      checkAndRecordImportBinding(importBinding, typesBySimpleNames, importReference, compoundName);
    } 
    if ((mask & 0x8) != 0 && importBinding.kind() == 8) {
      mask &= 0xFFFFFFF7;
      checkMoreStaticBindings(compoundName, typesBySimpleNames, mask, importReference);
    } 
  }
  
  private int checkAndRecordImportBinding(Binding importBinding, HashtableOfType typesBySimpleNames, ImportReference importReference, char[][] compoundName) {
    ReferenceBinding conflictingType = null;
    if (importBinding instanceof MethodBinding) {
      conflictingType = (ReferenceBinding)getType(compoundName, compoundName.length);
      if (!conflictingType.isValidBinding() || (importReference.isStatic() && !conflictingType.isStatic()))
        conflictingType = null; 
    } 
    char[] name = compoundName[compoundName.length - 1];
    if (importBinding instanceof ReferenceBinding || conflictingType != null) {
      ReferenceBinding referenceBinding = (conflictingType == null) ? (ReferenceBinding)importBinding : conflictingType;
      ReferenceBinding typeToCheck = (referenceBinding.problemId() == 3) ? 
        ((ProblemReferenceBinding)referenceBinding).closestMatch : 
        referenceBinding;
      if (importReference.isTypeUseDeprecated((TypeBinding)typeToCheck, this))
        problemReporter().deprecatedType((TypeBinding)typeToCheck, (ASTNode)importReference); 
      ReferenceBinding existingType = typesBySimpleNames.get(name);
      if (existingType != null) {
        if (TypeBinding.equalsEquals((TypeBinding)existingType, (TypeBinding)referenceBinding)) {
          for (int i = 0; i < this.importPtr; i++) {
            ImportBinding resolved = this.tempImports[i];
            if (resolved instanceof ImportConflictBinding) {
              ImportConflictBinding importConflictBinding = (ImportConflictBinding)resolved;
              if (TypeBinding.equalsEquals((TypeBinding)importConflictBinding.conflictingTypeBinding, (TypeBinding)referenceBinding) && 
                !importReference.isStatic()) {
                problemReporter().duplicateImport(importReference);
                recordImportBinding(new ImportBinding(compoundName, false, importBinding, importReference));
              } 
            } else if (resolved.resolvedImport == referenceBinding && 
              importReference.isStatic() != resolved.isStatic()) {
              problemReporter().duplicateImport(importReference);
              recordImportBinding(new ImportBinding(compoundName, false, importBinding, importReference));
            } 
          } 
          return -1;
        } 
        int j;
        int length;
        for (j = 0, length = this.topLevelTypes.length; j < length; j++) {
          if (CharOperation.equals((this.topLevelTypes[j]).sourceName, existingType.sourceName)) {
            problemReporter().conflictingImport(importReference);
            return -1;
          } 
        } 
        if (importReference.isStatic() && importBinding instanceof ReferenceBinding && (compilerOptions()).sourceLevel >= 3407872L)
          for (j = 0; j < this.importPtr; j++) {
            ImportBinding resolved = this.tempImports[j];
            if (resolved.isStatic() && resolved.resolvedImport instanceof ReferenceBinding && importBinding != resolved.resolvedImport && 
              CharOperation.equals(name, resolved.compoundName[resolved.compoundName.length - 1])) {
              ReferenceBinding type = (ReferenceBinding)resolved.resolvedImport;
              resolved.resolvedImport = (Binding)new ProblemReferenceBinding(new char[][] { name }, type, 3);
              return -1;
            } 
          }  
        problemReporter().duplicateImport(importReference);
        return -1;
      } 
      typesBySimpleNames.put(name, referenceBinding);
    } else if (importBinding instanceof FieldBinding) {
      for (int j = 0; j < this.importPtr; j++) {
        ImportBinding resolved = this.tempImports[j];
        if (resolved.isStatic() && resolved.resolvedImport instanceof FieldBinding && importBinding != resolved.resolvedImport && 
          CharOperation.equals(name, resolved.compoundName[resolved.compoundName.length - 1])) {
          if ((compilerOptions()).sourceLevel >= 3407872L) {
            FieldBinding field = (FieldBinding)resolved.resolvedImport;
            resolved.resolvedImport = (Binding)new ProblemFieldBinding(field, field.declaringClass, name, 3);
            return -1;
          } 
          problemReporter().duplicateImport(importReference);
          return -1;
        } 
      } 
    } 
    if (conflictingType == null) {
      recordImportBinding(new ImportBinding(compoundName, false, importBinding, importReference));
    } else {
      recordImportBinding((ImportBinding)new ImportConflictBinding(compoundName, importBinding, conflictingType, importReference));
    } 
    return this.importPtr;
  }
  
  public boolean hasDefaultNullnessFor(int location) {
    if (this.fPackage != null)
      return ((this.fPackage.defaultNullness & location) != 0); 
    return false;
  }
  
  public void registerInferredInvocation(Invocation invocation) {
    if (this.inferredInvocations == null)
      this.inferredInvocations = new ArrayList<>(); 
    this.inferredInvocations.add(invocation);
  }
  
  public void cleanUpInferenceContexts() {
    if (this.inferredInvocations == null)
      return; 
    for (Invocation invocation : this.inferredInvocations)
      invocation.cleanUpInferenceContexts(); 
    this.inferredInvocations = null;
  }
}
